#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from typing import *

from easypaint import EasyPaint


class Demo7(EasyPaint):
    x2: Optional[float] = None
    y2: Optional[float] = None
    lines_ids: List[Any] = []

    def on_mouse_button(self, button, x, y):
        if button == 2:
            self.close()

    def on_mouse_release(self, button, x, y):
        if button == 1:
            self.x2 = self.y2 = None

    def on_mouse_motion(self, button, x, y):
        if button == 1:
            if self.x2 is not None:
                self.lines_ids.append(self.create_line(self.x2, self.y2, x, y))
                while len(self.lines_ids) > 200:
                    self.erase(self.lines_ids[0])
                    del self.lines_ids[0]
            self.x2, self.y2 = x, y

    def main(self):
        self.easypaint_configure(title='Demo 7 - Dibujando con el ratón',
                                 background='white',
                                 size=(600, 600),
                                 coordinates=(0, 0, 1000, 1000))
        self.create_text(500, 50, "Dibuja con el botón izquierdo.", 8, 's')
        self.create_text(500, 0, "Termina programa pulsando el boton derecho.", 8, 's')


Demo7().run()



'''
Created on 22/10/2021

@author: David Llorens
@contact: dllorens@uji.es
@copyright: Universitat Jaume I de Castelló (2021)
@licence: GNU Affero General Public License v3
'''

# Descripción: Módulo para dibujar objetos 3D simples en EasyPaint.
#              Todavía muy optimizable.
#              El algoritmo de ocultación de lineas es la versión
#              Python del que aparece en el libro 'Gráficos por
#              Computador' de Ian O. Angell.
# Versión: 0.2 (beta)

import copy
import types
from math import cos, sin, atan2, sqrt

epsilon = 0.000001
epsilon1 = 1 + epsilon

OCULTAR_NO = 0
OCULTAR_SI = 1
OCULTAR_CARAS_TRASERAS = 2


class Escena3D:
    def __init__(self, ppd=600, ocultar=OCULTAR_SI, verb=0):
        self.o = []
        self.lin = []
        self.ppd = ppd
        self.dist = 0.0
        self.ocultar = ocultar
        self.verb = verb

    def insertar(self, o):
        self.o.append(o)

    def transforma3d(self, m):
        [obj.transforma3d(m) for obj in self.o]

    def puntoVista(self, pv, ppd=None):
        if ppd != None: self.ppd = ppd
        x, y, z = pv.x, pv.y, pv.z
        m = Matriz3D()
        m.traslacion(-x, -y, -z)
        rho = atan2(x, z)
        m.rotacionY(rho)
        phi = atan2(y, sqrt(x * x + z * z))
        m.rotacionX(-phi)
        self.transforma3d(m)

    def puntoVistaObj(self, pv, obj, ppd=None):
        if ppd != None: self.ppd = ppd
        x, y, z = pv.x, pv.y, pv.z
        m = Matriz3D()
        m.traslacion(-x, -y, -z)
        rho = atan2(x, z)
        m.rotacionY(rho)
        phi = atan2(y, sqrt(x * x + z * z))
        m.rotacionX(-phi)
        obj.transforma3d(m)

    def dibuja(self):
        a = len(self.o)
        self.lin = []
        verb = self.verb
        if self.ocultar == OCULTAR_SI:
            if verb:
                no = 1
                print("\nverb: Obteniendo caras visibles:")
            for obj in self.o:
                if verb:
                    print("verb: Objeto %d de %d" % (no, a))
                    no += 1
                obj.averiguarCarasVisibles(self.verb)
                obj.proyectarSiVisible(self.ppd, self.dist)
            if verb:
                no = 1
                print("\nverb: Obteniendo segmentos de línea visibles:")
            for obj in self.o:
                if verb:
                    print("verb: Objeto %d de %d" % (no, a))
                    no += 1
                lineas = obj.lineasR[1:]
                nl = 1
                old = -1
                for milinea in lineas:
                    if verb:
                        new = int(nl * 100.0 / len(lineas))
                        if new % 10 == 0 and new != old:
                            print("verb:     líneas %d%%" % (new))
                        nl += 1
                        old = new
                    pp1, pp2 = milinea
                    p1 = obj.puntos3dT[pp1]
                    p2 = obj.puntos3dT[pp2]
                    if not (p1.visible and p2.visible): continue
                    seg = self._oculta(p1, p2, milinea, obj)
                    list(map(self.lin.append, seg))
        elif self.ocultar == OCULTAR_CARAS_TRASERAS:
            if verb:
                no = 1
                print("\nverb: Obteniendo caras visibles:")
            for obj in self.o:
                if verb:
                    print("verb: Objeto %d de %d" % (no, a))
                    no += 1
                obj.averiguarCarasVisibles(self.verb)
                obj.proyectarSiVisible(self.ppd, self.dist)
            no = 1
            for obj in self.o:
                if verb:
                    print("verb: Objeto %d de %d" % (no, a))
                    no += 1
                # obj.proyectar(self.ppd,self.dist)
                lineas = obj.lineasR[1:]
                nl = 0
                old = -1
                for pp1, pp2 in lineas:
                    if verb:
                        new = int(nl * 100.0 / len(lineas) + 0.5)
                        if new % 10 == 0 and old != new:
                            print("verb:     líneas %d%%" % (new))
                        old = new
                        nl += 1
                    try:
                        x1, y1 = obj.puntos3dT[pp1].proy
                        x2, y2 = obj.puntos3dT[pp2].proy
                        self.lin.append((x1, y1, x2, y2))
                    except:
                        pass
        else:
            no = 1
            for obj in self.o:
                if verb:
                    print("verb: Objeto %d de %d" % (no, a))
                    no += 1
                obj.proyectar(self.ppd, self.dist)
                lineas = obj.lineasR[1:]
                nl = 0
                old = -1
                for pp1, pp2 in lineas:
                    if verb:
                        new = int(nl * 100.0 / len(lineas) + 0.5)
                        if new % 10 == 0 and old != new:
                            print("verb:     líneas %d%%" % (new))
                        old = new
                        nl += 1
                    x1, y1 = obj.puntos3dT[pp1].proy
                    x2, y2 = obj.puntos3dT[pp2].proy
                    self.lin.append((x1, y1, x2, y2))
        return self.lin

    def _oculta(self, p1, p2, mi_linea, mi_obj):
        x1, y1 = p1.proy
        x2, y2 = p2.proy
        rm = [[0.0, 1.0]]
        xd, yd = x2 - x1, y2 - y1
        for obj in self.o:
            l = list(range(len(obj.carasP)))
            for i in l:
                if obj.carasP[i][1] == 0: continue  # mira si es visible
                ok = 0
                if obj is mi_obj:
                    cx = obj.caras[i]
                    for x in cx:
                        if mi_linea == obj.lineasR[abs(x)]:
                            ok = 1
                            break
                    if ok: continue
                c = obj.carasP[i][0]
                cc = c + [c[0]]
                # Posibilidad A)
                xx1, yy1 = obj.puntos3dT[c[0]].proy
                fval = (yy1 - y1) * xd - (xx1 - x1) * yd
                if fval > 0:
                    fval = 1
                elif fval < 0:
                    fval = -1
                else:
                    fval = 0
                if fval != 0:
                    ok = 1
                    for p in cc[1:]:
                        xx1, yy1 = obj.puntos3dT[p].proy
                        fval2 = (yy1 - y1) * xd - (xx1 - x1) * yd
                        if fval2 > 0:
                            fval2 = 1
                        elif fval2 < 0:
                            fval2 = -1
                        else:
                            fval2 = 0
                        if fval - fval2 != 0:
                            ok = 0
                            break
                    if ok: continue
                # Posibilidad B)
                gval = (y2 - y1) * yd + (x2 - x1) * xd
                if gval > 0:
                    gval = 1
                elif gval < 0:
                    gval = -1
                else:
                    gval = 0
                if gval != 0:
                    ok = 1
                    for p in cc:
                        xx1, yy1 = obj.puntos3dT[p].proy
                        gval2 = (yy1 - y1) * yd + (xx1 - x1) * xd
                        if gval2 > 0:
                            gval2 = 1
                        elif gval2 < 0:
                            gval2 = -1
                        else:
                            gval2 = 0
                        if abs(gval - gval2) < 1.1:
                            ok = 0
                            break
                    if ok: continue
                # Posibilidad C)
                hval = (y1 - y2) * yd + (x1 - x2) * xd
                if hval > 0:
                    hval = 1
                elif hval < 0:
                    hval = -1
                else:
                    hval = 0
                if hval != 0:
                    ok = 1
                    for p in cc:
                        xx1, yy1 = obj.puntos3dT[p].proy
                        hval2 = (yy1 - y2) * yd + (xx1 - x2) * xd
                        if hval2 > 0:
                            hval2 = 1
                        elif hval2 < 0:
                            hval2 = -1
                        else:
                            hval2 = 0
                        if abs(hval - hval2) < 1.1:
                            ok = 0
                            break
                    if ok: continue
                #
                # Averiguar dos puntos de interseccion
                #
                rmax, rmin = 0.0, 1.0
                xx2, yy2 = obj.puntos3dT[c[0]].proy
                for p in cc[1:]:
                    xx1, yy1 = xx2, yy2
                    xx2, yy2 = obj.puntos3dT[p].proy
                    xe, ye = xx1 - xx2, yy1 - yy2
                    xf, yf = xx1 - x1, yy1 - y1
                    disk = xd * ye - xe * yd
                    if abs(disk) < epsilon:
                        # Si linea solapa una linea de la cara, salir bucle cara
                        if abs(xd) > epsilon:
                            xsi = xf / xd
                            if abs(yf - xsi * yd) < epsilon:
                                ok = 1
                                break
                        elif abs(xf) < epsilon:
                            ok = 1
                            break
                    else:
                        xsi = (xd * yf - yd * xf) / disk
                        # Si linea evita la cara salir bucle cara
                        if xsi < -epsilon or xsi > epsilon1: continue
                        rmu = (ye * xf - xe * yf) / disk
                        rmax, rmin = max(rmax, rmu), min(rmin, rmu)
                if ok: continue
                #
                if rmin > 1.0 or rmax < 0.0: continue
                rmax, rmin = min(1.0, rmax), max(0.0, rmin)
                if rmax - rmin <= 0.0: continue
                # Averiguar xmid y ymid
                rmid = (rmax + rmin) * 0.5
                rxx = 1.0 - rmid
                xmid = rxx * x1 + rmid * x2
                ymid = rxx * y1 + rmid * y2
                # Averiguar xhat, yhat y zhat del valor de phi
                denom = self.ppd * (p2.x - p1.x) - xmid * (p2.z - p1.z)
                if abs(denom) < epsilon:
                    denom = self.ppd * (p2.y - p1.y) - ymid * (p2.z - p1.z)
                    phi = (ymid * (p1.z + self.dist) - self.ppd * p1.y) / denom
                else:
                    phi = (xmid * (p1.z + self.dist) - self.ppd * p1.x) / denom
                zhat = (1.0 - phi) * p1.z + phi * p2.z
                ddd = (zhat + self.dist) / self.ppd
                xhat = xmid * ddd
                yhat = -ymid * ddd
                # calcular coef. plano a.x+b.y+c.z=d
                p1b = obj.puntos3dT[c[0]]
                p2b = obj.puntos3dT[c[1]]
                p3b = obj.puntos3dT[c[2]]
                dx1 = p1b.x - p2b.x
                dy1 = p1b.y - p2b.y
                dz1 = p1b.z - p2b.z
                dz3 = p3b.z - p2b.z
                dy3 = p3b.y - p2b.y
                dx3 = p3b.x - p2b.x
                a = dy1 * dz3 - dy3 * dz1
                b = dz1 * dx3 - dz3 * dx1
                c = dx1 * dy3 - dx3 * dy1
                d = a * p1b.x + b * p1b.y + c * p1b.z
                f1 = a * xhat + b * yhat + c * zhat - d
                if abs(f1) < epsilon: continue
                if f1 > 0:
                    f1 = 1
                elif f1 < 0:
                    f1 = -1
                else:
                    f1 = 0
                f2 = -self.dist * c - d
                if f2 > 0:
                    f2 = 1
                elif f2 < 0:
                    f2 = -1
                else:
                    f2 = 0
                if abs(f1 - f2) <= 1.0: continue
                # parte de la linea está oculta
                lrm = len(rm)
                for ii in range(lrm):
                    r1, r2 = rm[ii]
                    if r1 > rmax or r2 < rmin:
                        continue
                    elif r1 >= rmin and r2 <= rmax:
                        rm[ii][0] = -1.0
                    elif r1 < rmin:
                        if r2 > rmax: rm.append([rmax, r2])
                        rm[ii][1] = rmin
                    else:
                        rm[ii][0] = rmax
                # limpiar rm
                rm2 = []
                for r1, r2 in rm:
                    if r1 >= 0.0: rm2.append([r1, r2])
                if len(rm2) == 0: return []
                rm = rm2
        lin = []
        for s in rm:
            r1 = s[0]
            r2 = 1.0 - r1
            xp1 = x1 * r2 + x2 * r1
            yp1 = y1 * r2 + y2 * r1
            r1 = s[1]
            r2 = 1.0 - r1
            xp2 = x1 * r2 + x2 * r1
            yp2 = y1 * r2 + y2 * r1
            if (xp1 - xp2) != 0.0 and (yp1 - yp2) != 0.0:
                lin.append((xp1, yp1, xp2, yp2))
        return lin


class Objeto3D:
    def __init__(self, puntos3d, caras=[], carasP=[], lineas={}, lineasR=[0]):
        self.puntos3d = copy.deepcopy(puntos3d)
        self.puntos3dT = []  # [0.0]*len(self.puntos3d)
        self.caras = copy.deepcopy(caras)
        self.carasP = copy.deepcopy(carasP)
        self.lineas = copy.deepcopy(lineas)
        self.lineasR = copy.deepcopy(lineasR)
        self.lin = []
        self.numLin = 1  # tiene 1 de mas
        if carasP == [] or lineas == [] or lineasR == [0]:
            for c in caras: self.nuevaCara(c)

    def setup(self):
        self.puntos3d = copy.deepcopy(self.puntos3dT)

    def copia(self):
        return Objeto3D(self.puntos3d, self.caras, self.carasP,
                        self.lineas, self.lineasR)

    def nuevaCara(self, lInd):
        self.carasP.append([lInd[:], 1])
        cara = []
        aux = lInd[1:] + [lInd[0]]
        pOld = lInd[0]
        for p in aux:
            try:
                lin = self.lineas[(pOld, p)]
            except:
                try:
                    lin = -self.lineas[(p, pOld)]
                except:
                    self.lineas[(pOld, p)] = lin = self.numLin
                    self.lineasR.append((pOld, p))
                    self.numLin += 1
            cara.append(lin)
            pOld = p
        self.caras.append(cara)

    def transforma3d(self, m):
        self.puntos3dT = []
        aux = self.puntos3dT.append
        for punto in self.puntos3d:
            aux(m * punto)
        # for i in range(len(self.puntos3d)):
        #    self.puntos3dT[i] = m*self.puntos3d[i]

    def proyectar(self, ppd, dist=0.0):
        for p3d in self.puntos3dT:
            dd = ppd / (p3d.z + dist)
            p3d.proy = (p3d.x * dd, -p3d.y * dd)

    def proyectarSiVisible(self, ppd, dist=0.0):
        L = iter([p for p in self.puntos3dT if p.visible])
        for p3d in L:
            dd = ppd / (p3d.z + dist)
            p3d.proy = (p3d.x * dd, -p3d.y * dd)

    def averiguarCarasVisibles(self, verb=0):
        aux = self.puntos3dT
        for p in self.puntos3dT: p.visible = 0
        nc = 0
        old = -1
        for cc in self.carasP:
            if verb:
                new = int(nc * 100.0 / len(self.carasP))
                if new % 10 == 0 and new != old:
                    print("verb:     caras %d%%" % new)
                old = new
                nc += 1
            c = cc[0]
            p = aux[c[1]]
            v1 = aux[c[0]] - p  # de 1 a 0
            # normaliza v1
            longx = sqrt(v1.x * v1.x + v1.y * v1.y + v1.z * v1.z)
            v1.x /= longx;
            v1.y /= longx;
            v1.z /= longx

            v2 = aux[c[2]] - p  # de 1 a 2
            # normaliza v2
            longx = sqrt(v2.x * v2.x + v2.y * v2.y + v2.z * v2.z)
            v2.x /= longx;
            v2.y /= longx;
            v2.z /= longx

            if (v1 ** v2) * p > 0:  # cara invisible
                cc[1] = 0
            else:  # cara visible
                cc[1] = 1
                for pp in c:  # marcamos puntos visibles
                    aux[pp].visible = 1


class Punto3D:
    def __init__(self, x, y, z):
        if type(x) == types.FunctionType:
            self.fx = x
            self.fy = y
            self.fz = z
            self.posParametrica(0.0)
        else:
            self.x = x
            self.y = y
            self.z = z
            self.fx = self.fy = self.fz = None

    def __add__(self, otro):
        x = self.x + otro.x
        y = self.y + otro.y
        z = self.z + otro.z
        return Punto3D(x, y, z)

    def __sub__(self, otro):
        x = self.x - otro.x
        y = self.y - otro.y
        z = self.z - otro.z
        return Punto3D(x, y, z)

    def __pow__(self, otro):
        # <x1,y1,z1> X <x2,y2,z2> = <y1*z2 - z1*y2, z1*x2 - x1*z2, x1*y2 - y1*x2>
        x = self.y * otro.z - self.z * otro.y
        y = self.z * otro.x - self.x * otro.z
        z = self.x * otro.y - self.y * otro.x
        return Punto3D(x, y, z)

    def posParametrica(self, t):
        self.x = self.fx(t)
        self.y = self.fy(t)
        self.z = self.fz(t)

    def __mul__(self, otro):
        return (self.x * otro.x + self.y * otro.y + self.z * otro.z)

    def __str__(self):
        return '[%.2f %.2f %.2f]' % (self.x, self.y, self.z)

    def __repr__(self):
        return self.__str__()


class Matriz3D:
    def __init__(self, ini=None):
        if ini == None:
            self.mat = [[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 0.], [0., 0., 0., 1.]]
        else:
            self.mat = copy.deepcopy(ini)

    def puntoVista(self, pv):
        x, y, z = pv.x, pv.y, pv.z
        self.traslacion(-x, -y, -z)
        rho = atan2(x, z)
        self.rotacionY(rho)
        phi = atan2(y, sqrt(x * x + z * z))
        self.rotacionX(-phi)

    def identidad(self):
        self.mat = [[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 0.], [0., 0., 0., 1.]]

    def rotacionX(self, a):
        ca = cos(a)
        sa = sin(a)
        res = Matriz3D(self.mat)
        m0, m1, m2, m3 = self.mat
        r0, r1, r2, r3 = res.mat
        r0[1] = m0[1] * ca + m0[2] * sa
        r0[2] = m0[2] * ca - m0[1] * sa
        r1[1] = m1[1] * ca + m1[2] * sa
        r1[2] = m1[2] * ca - m1[1] * sa
        r2[1] = m2[1] * ca + m2[2] * sa
        r2[2] = m2[2] * ca - m2[1] * sa
        r3[1] = m3[1] * ca + m3[2] * sa
        r3[2] = m3[2] * ca - m3[1] * sa
        self.mat = res.mat

    def rotacionY(self, a):
        res = Matriz3D(self.mat)
        ca = cos(a)
        sa = sin(a)
        m0, m1, m2, m3 = self.mat
        r0, r1, r2, r3 = res.mat
        r0[0] = m0[0] * ca - m0[2] * sa
        r0[2] = m0[0] * sa + m0[2] * ca
        r1[0] = m1[0] * ca - m1[2] * sa
        r1[2] = m1[0] * sa + m1[2] * ca
        r2[0] = m2[0] * ca - m2[2] * sa
        r2[2] = m2[0] * sa + m2[2] * ca
        r3[0] = m3[0] * ca - m3[2] * sa
        r3[2] = m3[0] * sa + m3[2] * ca
        self.mat = res.mat

    def rotacionZ(self, a):
        res = Matriz3D(self.mat)
        ca = cos(a)
        sa = sin(a)
        m0, m1, m2, m3 = self.mat
        r0, r1, r2, r3 = res.mat
        r0[0] = m0[0] * ca + m0[1] * sa
        r0[1] = m0[1] * ca - m0[0] * sa
        r1[0] = m1[0] * ca + m1[1] * sa
        r1[1] = m1[1] * ca - m1[0] * sa
        r2[0] = m2[0] * ca + m2[1] * sa
        r2[1] = m2[1] * ca - m2[0] * sa
        r3[0] = m3[0] * ca + m3[1] * sa
        r3[1] = m3[1] * ca - m3[0] * sa
        self.mat = res.mat

    def escalado(self, x, y, z):
        rot = Matriz3D([[x, 0.0, 0.0, 0.0],
                        [0.0, y, 0.0, 0.0],
                        [0.0, 0.0, z, 0.0],
                        [0.0, 0.0, 0.0, 1.0]])
        self.mat = self.__mul__(rot).mat

    def traslacion(self, x, y, z):
        m0, m1, m2, m3 = self.mat
        m0[0] += m0[3] * x
        m0[1] += m0[3] * y
        m0[2] += m0[3] * z
        m1[0] += m1[3] * x
        m1[1] += m1[3] * y
        m1[2] += m1[3] * z
        m2[0] += m2[3] * x
        m2[1] += m2[3] * y
        m2[2] += m2[3] * z
        m3[0] += m3[3] * x
        m3[1] += m3[3] * y
        m3[2] += m3[3] * z

    def demo(self):
        for i in [0, 1, 2, 3]:
            for j in [0, 1, 2, 3]:
                self.mat[i][j] = float(i * 4 + j + 1)

    def copia(self):
        return Matriz3D(self.mat)

    def __str__(self):
        res = ''
        for i in [0, 1, 2, 3]:
            res += '|'
            for j in [0, 1, 2, 3]:
                res += ' %6.2f' % self.mat[i][j]
            res = res + ' |\n'
        return res

    def __add__(self, otro):
        mat = Matriz3D()
        print(type(otro))
        for i in [0, 1, 2, 3]:
            for j in [0, 1, 2, 3]:
                mat[i][j] = self.mat[i][j] + otro.mat[i][j]
        return mat

    def __mul__(self, otro):
        m0, m1, m2, m3 = self.mat
        try:
            o0, o1, o2, o3 = otro.mat
            res = Matriz3D()
            r0, r1, r2, r3 = res.mat

            r0[0] = m0[0] * o0[0] + m0[1] * o1[0] + m0[2] * o2[0] + m0[3] * o3[0]
            r0[1] = m0[0] * o0[1] + m0[1] * o1[1] + m0[2] * o2[1] + m0[3] * o3[1]
            r0[2] = m0[0] * o0[2] + m0[1] * o1[2] + m0[2] * o2[2] + m0[3] * o3[2]
            r0[3] = m0[0] * o0[3] + m0[1] * o1[3] + m0[2] * o2[3] + m0[3] * o3[3]

            r1[0] = m1[0] * o0[0] + m1[1] * o1[0] + m1[2] * o2[0] + m1[3] * o3[0]
            r1[1] = m1[0] * o0[1] + m1[1] * o1[1] + m1[2] * o2[1] + m1[3] * o3[1]
            r1[2] = m1[0] * o0[2] + m1[1] * o1[2] + m1[2] * o2[2] + m1[3] * o3[2]
            r1[3] = m1[0] * o0[3] + m1[1] * o1[3] + m1[2] * o2[3] + m1[3] * o3[3]

            r2[0] = m2[0] * o0[0] + m2[1] * o1[0] + m2[2] * o2[0] + m2[3] * o3[0]
            r2[1] = m2[0] * o0[1] + m2[1] * o1[1] + m2[2] * o2[1] + m2[3] * o3[1]
            r2[2] = m2[0] * o0[2] + m2[1] * o1[2] + m2[2] * o2[2] + m2[3] * o3[2]
            r2[3] = m2[0] * o0[3] + m2[1] * o1[3] + m2[2] * o2[3] + m2[3] * o3[3]

            r3[0] = m3[0] * o0[0] + m3[1] * o1[0] + m3[2] * o2[0] + m3[3] * o3[0]
            r3[1] = m3[0] * o0[1] + m3[1] * o1[1] + m3[2] * o2[1] + m3[3] * o3[1]
            r3[2] = m3[0] * o0[2] + m3[1] * o1[2] + m3[2] * o2[2] + m3[3] * o3[2]
            r3[3] = m3[0] * o0[3] + m3[1] * o1[3] + m3[2] * o2[3] + m3[3] * o3[3]
            return res
        except:
            x = otro.x * m0[0] + otro.y * m1[0] + otro.z * m2[0] + m3[0]
            y = otro.x * m0[1] + otro.y * m1[1] + otro.z * m2[1] + m3[1]
            z = otro.x * m0[2] + otro.y * m1[2] + otro.z * m2[2] + m3[2]
            return Punto3D(x, y, z);


class Malla3D(Objeto3D):
    def __init__(self, mat):
        fil = len(mat)
        col = len(mat[0])
        self.lp = [Punto3D(f * 10, mat[f][c], c * 10) for f in range(fil) for c in range(col)]

        # m = Matriz3D()
        # m.escalado(1,1,1)
        # m.traslacion(0,0,0)
        # self.transforma3d(m)
        # self.setup()
        self.lc = []
        for f in range(fil - 1):
            for c in range(col - 1):
                # print(f,c)
                p = f * fil + c
                p2 = (f + 1) * fil + c
                self.lc.append([p2 + 1, p + 1, p])
                self.lc.append([p2, p2 + 1, p])

        Objeto3D.__init__(self, self.lp, caras=self.lc)
        m = Matriz3D()
        # m.escalado(1,1,1)
        # m.traslacion(0,0,0)
        self.transforma3d(m)
        self.setup()
        # list(map(self.nuevaCara,self.lc))


class Rectangulo3D(Objeto3D):
    lp = [Punto3D(1, -1, -1), Punto3D(1, -1, 1),
          Punto3D(1, 1, 1), Punto3D(1, 1, -1),
          Punto3D(-1, -1, -1), Punto3D(-1, -1, 1),
          Punto3D(-1, 1, 1), Punto3D(-1, 1, -1)]
    lc = [[0, 1, 2, 3], [7, 6, 5, 4], [3, 2, 6, 7],
          [0, 4, 5, 1], [1, 5, 6, 2], [0, 3, 7, 4]]

    def __init__(self, pTam, pPos):
        Objeto3D.__init__(self, self.lp)
        m = Matriz3D()
        m.escalado(pTam.x / 2, pTam.y / 2, pTam.z / 2)
        m.traslacion(pPos.x, pPos.y, pPos.z)
        self.transforma3d(m)
        self.setup()
        list(map(self.nuevaCara, self.lc))


class Piramide3D(Objeto3D):
    lp = [Punto3D(1, -1, 1), Punto3D(1, -1, -1),
          Punto3D(-1, -1, -1), Punto3D(-1, -1, 1),
          Punto3D(0, 1, 0)]
    lc = [[0, 1, 2, 3], [1, 0, 4], [2, 1, 4], [3, 2, 4], [0, 3, 4]]

    def __init__(self, pTam, pPos):
        Objeto3D.__init__(self, self.lp)
        m = Matriz3D()
        m.escalado(pTam.x / 2, pTam.y / 2, pTam.z / 2)
        m.traslacion(pPos.x, pPos.y, pPos.z)
        self.transforma3d(m)
        self.setup()
        list(map(self.nuevaCara, self.lc))


class Cubo3D(Rectangulo3D):
    def __init__(self, lado, p):
        Rectangulo3D.__init__(self, Punto3D(lado, lado, lado), p)


if __name__ == '__main__':
    m = Matriz3D()
    m.demo()
    m2 = m.copia()
    print(m)
    v = Punto3D(1, 2, 3)
    # c=m+m
    print(m)
    m.rotacionX(0.5)
    print(m)



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from math import sin, cos, pi

from easypaint import EasyPaint


def rueda(x0, y0, radio, num_dientes, prof_diente, ini):
    r1 = radio - prof_diente / 2.0
    r2 = radio + prof_diente / 2.0
    values = []
    da = 2.0 * pi / num_dientes / 4.0
    cte = 2.0 * pi / num_dientes
    ini = ini - da * 3 / 2
    for i in range(0, num_dientes):
        angle = i * cte + ini
        values.append((r1 * cos(angle) + x0, r1 * sin(angle) + y0))
        values.append((r2 * cos(angle + da) + x0, r2 * sin(angle + da) + y0))
        values.append((r2 * cos(angle + 2 * da) + x0, r2 * sin(angle + 2 * da) + y0))
        values.append((r1 * cos(angle + 3 * da) + x0, r1 * sin(angle + 3 * da) + y0))
    return values + [values[0]]


def mcm(i, j):
    while 1:
        r = i % j
        if r == 0: return j
        i, j = j, r


class Demo6(EasyPaint):
    rold = []
    # Cambiando estos valores se obtiene dibujos diferentes
    radioCirculo = 200
    radioDisco = 75
    distLapizCentroDisco = 50
    velocidad = 0.04  # reducir para aumentar velocidad
    dientesCirculo = 40  # no es necesario cambiar este valor

    def spyro(self, a, b, rd, di):
        def anim(cc):
            nonlocal ang, theta, x2, y2, xx2, yy2, o_id
            ang += 1
            theta += thetad
            phi = -ang * inc
            xx, yy = rab * cos(theta), rab * sin(theta)
            x, y = xx + rd * cos(phi), yy - rd * sin(phi)
            self.create_line(x, y, x2, y2)
            self.move('lapiz', x - x2, y - y2)
            c = rueda(xx, yy, b - prof_d / 2, n2, prof_d, -phi)
            self.dibuja_rueda(c, 'blue')
            self.move('disco', xx - xx2, yy - yy2)
            self.erase(o_id)
            o_id = self.create_line(x, y, xx, yy, 'blue')
            x2, y2 = x, y
            xx2, yy2 = xx, yy
            self.update()
            # time.sleep(self.velocidad)
            if cc > 0:
                self.after(int(self.velocidad * 1000), lambda: anim(cc - 1))

        prof_d = 10  # profundidad diente
        self.create_filled_circle(0, 0, 3, 'red')
        xx2, yy2 = a - b, 0
        print("Tamaño diente: %g" % (float(a) / di))
        print("Nº dientes agujero:", di)
        o_id = rueda(0, 0, a - prof_d / 2, di, prof_d, 0)
        self.dibuja_rueda(o_id, 'red')
        self.rold = []  # para que no se borre la rueda roja (es fija)
        n2 = int(di * float(b) / a)
        print("Nº dientes disco:", n2)
        b2 = n2 * float(a) / di
        if b2 != b:
            print("Ajustado radio del disco: %g -> %g" % (b, b2))
            b = b2

        rab = (a - b)
        o_id = rueda(rab, 0, b - prof_d / 2, n2, prof_d, 0)
        self.dibuja_rueda(o_id, 'blue')
        self.create_filled_circle(rab, 0, 3, 'blue', tags='disco')
        self.create_filled_circle(rab + rd, 0, 3, 'black', tags='lapiz')
        theta = 0.0
        thetad = pi * 0.02  # 12#52
        n = n2 / mcm(di, n2)
        print("Vueltas:", n)
        x2, y2 = rab + rd, 0.0
        ang = 0
        inc = -((di - n2) * (2 * pi / n2)) / 100
        o_id = self.create_line(x2, y2, xx2, yy2, 'blue')
        self.after(0, lambda: anim(int(100 * n)))

    def dibuja_rueda(self, o_id, color):
        r = []
        x2, y2 = o_id[0]
        for x, y in o_id[1:]:
            r.append(self.create_line(x, y, x2, y2, color))
            if len(self.rold) > 0:
                self.erase(self.rold[0])
                del self.rold[0]
            x2, y2 = x, y
        self.erase(self.rold)
        self.rold = r

    def on_key_press(self, keysym):
        self.close()

    def main(self):
        self.easypaint_configure(title='Demo 6 - Spyro',
                                 background='white',
                                 size=(600, 600),
                                 coordinates=(-250, -250, 250, 250))
        self.spyro(self.radioCirculo, self.radioDisco, self.distLapizCentroDisco, self.dientesCirculo)

        self.create_text(0, -250, "Press any key to exit", 12, 's')


Demo6().run()



#!/usr/bin/env python3
"""
Created on 30/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""

from math import pi
from typing import *

from easypaint import EasyPaint
from easypaint.libsimple3d import *


class SomaAnim(EasyPaint):
    ppd = 1000
    escena = Escena3D(ppd)

    def __init__(self, sol):
        EasyPaint.__init__(self)
        self.sol = sol
        self.cubos = {}

    def build(self):
        self.cubos = {}
        f = {}
        for x in range(3):
            for y in range(3):
                for z in range(3):
                    key = self.sol[x][y][z]
                    if key in f:
                        f[key].append((x, y, z))
                    else:
                        f[key] = [(x, y, z)]
        lado = 800
        for key in f:
            for cube in f[key]:
                p = Punto3D(cube[0] * lado, cube[1] * lado, cube[2] * lado)
                cubo = Cubo3D(lado * 0.9999999999, p)
                if key in self.cubos:
                    self.cubos[key].append(cubo)
                else:
                    self.cubos[key] = [cubo]
                m = Matriz3D()
                m.traslacion(-1 * lado, -1 * lado, -1 * lado)
                cubo.transforma3d(m)
                cubo.setup()
                self.escena.insertar(cubo)
        return f

    def on_key_press(self, keysym):
        if keysym in ['Return', 'Escape']:
            self.close()

    def main(self):
        def anim(step, max_step):
            nonlocal t, nl2, lin2, sc_old, ii
            if step < max_step // 2:
                ii = step
            else:
                ii = 100 - step

            pv.posParametrica(t)
            self.escena.puntoVista(pv)
            lineas = self.escena.dibuja()
            lin = []
            for i in range(len(lineas)):
                lin.append(self.create_line(*lineas[i]))
                if i < nl2: self.erase(lin2[i])
            self.erase(lin2[len(lin):])
            lin2 = lin[:]
            nl2 = len(lin)

            self.update()

            t = (t + inc) % pi2
            sc = 30 * ii
            for key in self.cubos:
                xm = sum([cubo[0] for cubo in f[key]]) / len(f[key])
                ym = sum([cubo[1] for cubo in f[key]]) / len(f[key])
                zm = sum([cubo[2] for cubo in f[key]]) / len(f[key])

                for cubo in self.cubos[key]:
                    m = Matriz3D()
                    if sc_old is not None:
                        m.traslacion(-(xm - 1) * sc_old, -(ym - 1) * sc_old, -(zm - 1) * sc_old)
                    m.traslacion((xm - 1) * sc, (ym - 1) * sc, (zm - 1) * sc)
                    cubo.transforma3d(m)
                    cubo.setup()
            sc_old = sc

            step = (step + 1) % max_step
            self.after(0, lambda: anim(step, max_step))

        self.easypaint_configure(title='Soma cube - Animated solution',
                                 background='lightblue',
                                 size=(501, 401),
                                 coordinates=(-500, -400, 500, 400))
        self.create_text(0, -390, "Press 'Return' or 'Escape' to exit", 10, 'S', 'blue')

        f = self.build()

        pi2 = 2 * pi
        lin2 = []
        nl2 = 0
        sc_old: Optional[float] = None
        t = 0.0000001

        pv = Punto3D(lambda tt: 8000 * sin(tt),
                     lambda tt: 5000,
                     lambda tt: 8000 * cos(tt))

        inc = 1.003 * pi / 100  # 0.042
        ii = 0
        self.after(0, lambda: anim(0, 100))


if __name__ == '__main__':
    sol1 = [[[1, 7, 7], [4, 4, 7], [3, 4, 4]], [[1, 5, 2], [1, 5, 7], [3, 3, 6]], [[2, 2, 2], [5, 5, 6], [3, 6, 6]]]
    sol2 = [[[3, 6, 6], [1, 5, 6], [1, 5, 4]], [[3, 3, 6], [1, 7, 4], [5, 5, 4]], [[3, 7, 7], [2, 7, 4], [2, 2, 2]]]
    sa = SomaAnim(sol1)
    sa.run()



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from enum import Enum

from easypaint import EasyPaint


class State(Enum):
    Initial = 1
    Delete = 2
    Exit = 3


class Demo4(EasyPaint):
    state: State = State.Initial
    textId = None

    def on_key_press(self, keysym):
        if self.state == State.Initial:
            pass
        elif self.state == State.Delete:
            self.erase(['c2', self.textId])
            self.create_text(500, 0, "Press any key to exit", 12, 's')
            self.state = State.Exit
        elif self.state == State.Exit:
            self.close()

    def animation(self, c):
        self.move('c', 1, 0)
        self.move('c2', -1, 0)
        if c > 0:
            self.after(10, lambda: self.animation(c - 1))
        else:
            self.textId = self.create_text(500, 0, "Press any key to delete left circles", 12, 's')
            self.state = State.Delete

    def main(self):
        self.easypaint_configure(title='Demo 4 - Interferencia',
                                 background='steelblue',
                                 size=(600, 600),
                                 coordinates=(0, 0, 1000, 1000))
        for i in range(1, 16):
            self.create_circle(300, 500, i * 20, 'black', tags='c')
            self.create_circle(700, 500, i * 20, 'black', tags='c2')

        self.after(0, lambda: self.animation(400))


Demo4().run()



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from easypaint import EasyPaint


class Demo1(EasyPaint):
    def on_key_press(self, keysym):
        self.close()

    def main(self):
        x1, y1, x2, y2 = (0, 0, 399, 399)
        self.easypaint_configure(title='Demo 0 - Window size and window coordinates',
                                 background='white',
                                 size=(400, 400),
                                 coordinates=(x1, y1, x2, y2))
        print(self.scale, self.size, self.coordinates)

        c = 'red'
        self.create_line(x1, y1, x1, y2, c)
        self.create_line(x1, y1, x2, y1, c)
        self.create_line(x2, y2, x2, y1, c)
        self.create_line(x2, y2, x1, y2, c)

        # escribe texto
        self.create_text((x1 + x2) / 2, (y1 + y2) / 2, "Press any key to exit", 12, 'c')


Demo1().run()



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from easypaint import EasyPaint


class Demo8(EasyPaint):
    def __init__(self):
        super().__init__()
        self.ls = []

    def on_mouse_button(self, button, x, y):
        if button == 1:
            self.ls.append(self.create_circle(x, y, 20))
            if len(self.ls) > 20:  # como máximo se permiten 20 círculos
                self.erase(self.ls[0])
                del self.ls[0]
        elif button == 2:
            self.close()

    def main(self):
        self.easypaint_configure(title='Demo 8 - Uso de los botones del ratón',
                                 background='white',
                                 size=(600, 600),
                                 coordinates=(0, 0, 599, 599))
        self.create_text(300, 290, "Botón izq.: dibuja círculo", 8, 'c')
        self.create_text(300, 310, "Botón der.: termina programa", 8, 'c')


Demo8().run()



#!/usr/bin/env python3
"""
Created on 30/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from easypaint import EasyPaint


class Demo9(EasyPaint):
    id = None

    def on_key_release(self, keysym):
        if self.id is not None:
            self.erase(self.id)

    def on_key_press(self, keysym):
        if self.id is not None:
            self.erase(self.id)
        self.id = self.create_text(200, 100, f"{keysym}", 40, 'c')
        if keysym == 'Escape':
            self.close()

    def main(self):
        self.easypaint_configure(title='Demo 9 - readkey',
                                 background='white',
                                 size=(401, 201),
                                 coordinates=(0, 0, 400, 200))
        self.create_text(200, 0, "Press any key. 'Escape' to exit.", 10, 's')


Demo9().run()



"""
2024/09/30 - 1.0.7 - Corregido bug cuando self.coordinates es None.
2024/09/27 - 1.0.6 - Corregido __slots__ y tipo del método after.
                   - Added 'TVERSION' constant
                   - Corrige bug en 'FIT' (ver 1.0.5).
2023/09/21 - 1.0.5 - Added 'VERSION' constant
                   - easypaint_configure(...): Parameter 'size' now has three options:
                        'FIT'   : max available size keaping aspect ratio (see param 'coordinates')
                        'FULL'  : max available size
                        (width, height)
2022/09/22 - 1.0.4 - Corregido create_filled_polygon(...)
2022/09/22 - 1.0.3 - Añade create_filled_polygon(...)
2022/09/22 - 1.0.2 - Añade create_polygon(...)
2022/09/01 - 1.0.1 - Biblioteca para crear ventanas con un lienzo.

@author: David Llorens
@contact: dllorens@uji.es
@copyright: Universitat Jaume I de Castelló (2024)
@licence: GNU Affero General Public License v3
"""

import tkinter
from abc import ABC, abstractmethod
from typing import Any, Callable

TVERSION = (1, 0, 7)
VERSION = '.'.join([str(e) for e in TVERSION])

DEFAULT_CANVAS_WIDTH = 500
DEFAULT_CANVAS_HEIGHT = 500

class EasyPaintException(Exception):
    def __init__(self, message, value=None):
        self.value = value
        self.message = message
        super().__init__(self.message)


class EasyPaint(ABC):
    __slots__ = ('_title', '_background', '_root', '_canvas',
                 '_width', '_height', '_xscale', '_yscale',
                 '_left', '_right', '_top', '_bottom', '_screensize', 'closing')

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        self._root.title(self._title)

    @property
    def size(self):
        return self._width, self._height

    @size.setter
    def size(self, value: tuple[int, int] | str):
        msg = "easypaint_configure: Parameter 'size' must be a tuple of two integers greater than 0 or 'FIT' or 'FULL'"
        if isinstance(value, str):
            if value == 'FULL':
                size_t = self._screensize
            elif value == 'FIT':
                l, b, r, t = self.coordinates
                ar = abs(r - l)/abs(t - b)

                size_t = self._screensize
                ar2 = size_t[0]/size_t[1]
                if ar < ar2:
                    size_t = int(ar*size_t[1]), size_t[1]
                elif ar > ar2:
                    size_t = size_t[0], int(size_t[0]/ar)
            else:
                raise EasyPaintException(msg)
        elif (isinstance(value, tuple) and len(value) == 2 and isinstance(value[0], int) and
              isinstance(value[1], int) and value[0] > 0 and value[1] > 0):
            size_t = value
        else:
            raise EasyPaintException(msg)

        self._width, self._height = size_t
        if hasattr(self, '_canvas'):
            self.erase()
            canvas_args = {"width": min(self._width, self._screensize[0]),
                           "height": min(self._height, self._screensize[1])}
            self._canvas.configure(**canvas_args)
        self._set_scale()

    @property
    def coordinates(self):
        return self._left, self._bottom, self._right, self._top

    @coordinates.setter
    def coordinates(self, value: tuple[float, float, float, float]):
        if hasattr(self, '_canvas'):
            self.erase()
        self._left, self._bottom, self._right, self._top = value
        self._set_scale()

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, value):
        self._background = value
        if hasattr(self, '_canvas'):
            canvas_args = {"background": value if value is not None else 'white'}
            self._canvas.configure(**canvas_args)
            if value is None:
                self._set_transparent_background()

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def top(self):
        return self._top

    @property
    def bottom(self):
        return self._bottom

    @property
    def scale(self) -> tuple[float, float]:
        return self._xscale, self._yscale

    @property
    def center(self) -> tuple[float, float]:
        return (self._right + self._left) / 2, (self._top + self._bottom) / 2

    def __init__(self):
        self.closing = False

        self.background = 'white'

        self._width, self._height = DEFAULT_CANVAS_WIDTH, DEFAULT_CANVAS_HEIGHT
        self._left, self._bottom, self._right, self._top = 0, 0, self._width - 1, self._height - 1
        self._set_scale()

        self._root = tkinter.Tk()
        self._screensize = self._root.maxsize()
        self.title = 'EasyPaint'
        self._root.resizable(width=False, height=False)
        self._root.protocol("WM_DELETE_WINDOW", self.close)
        self._root.bind('<KeyPress>', lambda e: self.on_key_press(e.keysym))
        self._root.bind('<KeyRelease>', lambda e: self.on_key_release(e.keysym))

        self._canvas = tkinter.Canvas(self._root, borderwidth=0, highlightthickness=0,
                                      height=self._height, width=self._width, background=self.background)
        self._canvas.pack(padx=0, pady=0)
        self._canvas.bind('<Button>', self._on_mouse_button)
        self._canvas.bind('<ButtonRelease>', self._on_mouse_release)
        self._canvas.bind('<B1-Motion>', lambda e: self._on_mouse_motion(1, e))
        self._canvas.bind('<B2-Motion>', lambda e: self._on_mouse_motion(2, e))
        self._canvas.bind('<B3-Motion>', lambda e: self._on_mouse_motion(3, e))
        self._canvas.bind('<Leave>', self.on_mouse_leave)

    def easypaint_configure(self, size: tuple[int, int] | str = 'FIT',
                            coordinates: tuple[float, float, float, float] | None = None,
                            title: str | None = 'EasyPaint',
                            background: str | None = 'white'):
        """Configure the window

        Arguments:
            size -- 'FIT'   : max available size keaping aspect ratio (see param 'coordinates')
                    'FULL'  : max available size
                    (width, height)

            coordinates -- (left, bottom, right, top). Default is (0, 0, width-1, height-1)

            title -- window title. Default is 'EasyPaint'

            background -- color name. Default is 'white'
        """
        if self.closing: return

        self.erase()

        if coordinates is None:
            try:
                self._left, self._bottom, self._right, self._top = 0, 0, size[0]-1, size[1]-1
            except:
                self._left, self._bottom, self._right, self._top = 0, 0, 1000, 1000
        else:
            self._left, self._bottom, self._right, self._top = coordinates

        self.size = size
        self._set_scale()

        if coordinates is not None:
            self.coordinates = coordinates

        self.background = background

        if title is not None:
            self.title = title

    # PRIVATE METHODS ----------------------------------------------------------------

    def _set_scale(self):
        ww = self._right - self._left
        iw = 1 if ww >= 0 else -1
        self._xscale = self._width / float(ww + iw)
        hh = self._top - self._bottom
        ih = 1 if hh >= 0 else -1
        self._yscale = self._height / float(hh + ih)

    def _set_transparent_background(self):
        if self.closing: return
        self._root.wm_attributes("-transparent", True)
        self._root.config(bg='systemTransparent')
        self._canvas.config(bg='systemTransparent')

    def _transform(self, x, y):
        xb = int((x - self._left) * self._xscale)
        yb = int((self._top - y) * self._yscale)
        return xb, yb

    def _transform_x(self, x, ):
        return int((x - self._left) * self._xscale)

    def _transform_y(self, y):
        return int((self._top - y) * self._yscale)

    def _on_mouse_release(self, event):
        if self.closing: return
        x = event.x / self._xscale + self._left
        y = self._top - event.y / self._yscale
        self.on_mouse_release(event.num, x, y)

    def _on_mouse_button(self, event):
        if self.closing: return
        x = event.x / self._xscale + self._left
        y = self._top - event.y / self._yscale
        self.on_mouse_button(event.num, x, y)

    def _on_mouse_motion(self, button, event):
        if self.closing: return
        x = event.x / self._xscale + self._left
        y = self._top - event.y / self._yscale
        self.on_mouse_motion(button, x, y)

    # -----------------------------------------------------------------

    def on_mouse_release(self, button, x, y):
        pass

    def on_mouse_button(self, button, x, y):
        pass

    def on_mouse_motion(self, button, x, y):
        pass

    def on_mouse_leave(self, event):
        pass

    def on_key_press(self, keysym):
        pass

    def on_key_release(self, keysym):
        pass

    # -----------------------------------------------------------------

    def update(self):
        """Enter event loop until all pending events have been processed by Tcl.
        """
        if self.closing: return
        self._canvas.update()  # animaciones más suaves
        # self._canvas.update_idletasks()    # animaciones más bruscas y rapidas

    def create_rectangle(self, x1: float, y1: float, x2: float, y2: float, color: str = 'black', fill=None, **args):
        """Draws a rectangle

        Arguments:
            x1, y1 -- lower left point coordinates

            x2, y2 -- upper right point coordinates

            color -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the rectangle: erase(id)
        """
        if self.closing: return
        fn = 'create_rectangle' if fill is None else 'create_filled_rectangle'
        args['outline'] = color[:]
        if fill is not None: args['fill'] = fill[:]
        try:
            x1b, y1b = self._transform(x1, y1)
            x2b, y2b = self._transform(x2, y2)
            if x2b < x1b:
                x1b, x2b = x2b, x1b
            if y2b < y1b:
                y1b, y2b = y2b, y1b
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in {fn}: {(x1, y1, x2, y2)}")

        try:
            return self._canvas.create_rectangle(*(x1b, y1b, x2b, y2b), **args)
        except Exception as _e:
            raise EasyPaintException(f"{fn}")

    def create_filled_rectangle(self, x1: float, y1: float, x2: float, y2: float, color: str = 'black', fill=None,
                                **args):
        """Draws a filled rectangle

        Arguments:
            x1, y1 -- lower left point coordinates

            x2, y2 -- upper right point coordinates

            color -- color name (default is 'black')

            fill -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the rectangle: erase(id)
        """
        if self.closing: return
        if fill is None: fill = color
        return self.create_rectangle(x1, y1, x2, y2, color, fill, **args)

    def create_circle(self, x: float, y: float, radius: float, color: str = 'black', fill=None, **args):
        """Draws a circle

        Arguments:
            x, y -- point coordinates

            radius -- float

            color -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the circle: erase(id)
        """
        if self.closing: return
        fn = 'create_circle' if fill is None else 'create_filled_circle'
        args['outline'] = color[:]
        if fill is not None: args['fill'] = fill[:]
        try:
            x1b, y1b = self._transform(x - radius, y - radius)
            x2b, y2b = self._transform(x + radius, y + radius)
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in {fn}: {(x, y)}")
        try:
            return self._canvas.create_oval(*(x1b, y1b, x2b, y2b), **args)
        except Exception as _e:
            raise EasyPaintException(f"{fn}")

    def create_filled_circle(self, x: float, y: float, radius: float, color='black', fill=None, **args):
        """Draws a filled circle

        Arguments:
            x, y -- point coordinates

            radius -- float

            color -- color name (default is 'black')

            fill -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the circle: erase(id)
        """
        if self.closing: return
        if fill is None: fill = color
        return self.create_circle(x, y, radius, color, fill, **args)

    def create_polygon(self, *params, color: str = 'black', fill: str = None, **args):
        """Draws a polygon

        Arguments:
            x0, y0, x1, y1, ...  -- point coordinates

            color -- outline color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the polygon: erase(id)
        """
        if self.closing: return
        args['fill'] = '' if fill is None else fill
        args['outline'] = color
        try:
            params2 = [self._transform_x(e) if i % 2 == 0 else self._transform_y(e) for i, e in enumerate(params)]
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in create_polygon: {params}")
        try:
            return self._canvas.create_polygon(*params2, **args)
        except Exception as e:
            raise EasyPaintException(f"create_polygon: {e}")

    def create_filled_polygon(self, *params, color: str = 'black', fill: str = None, **args):
        """Draws a filled polygon

        Arguments:
            x0, y0, x1, y1, ...  -- point coordinates

            color -- outline color name (default is 'black')

            fill -- color name (default is same as color)

        Returns:
            A number (identifier). You can use this 'id' to move or delete the polygon: erase(id)
        """
        if fill is None:
            fill = color
        if self.closing: return
        return self.create_polygon(*params, color=color, fill=fill, **args)

    def create_point(self, x: float, y: float, color: str = 'black', **args):
        """Draws a point

        Arguments:
            x, y -- point coordinates

            color -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move o delete the point: erase(id)
        """
        if self.closing: return
        args['fill'] = color[:]
        args['width'] = 2
        try:
            x1b, y1b = self._transform(x, y)
            x2b = x1b + 2
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in create_point: {(x, y)}")
        try:
            return self._canvas.create_line(*(x1b, y1b, x2b, y1b), **args)
        except Exception as _e:
            raise EasyPaintException("create_point")

    def create_line(self, x1: float, y1: float, x2: float, y2: float, color: str = 'black', **args):
        """Draws a line between two points

        Arguments:
            x1, y1 -- start point coordinates
            x2, y2 -- end point coordinates
            color -- color name (default is 'black')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the line
        """
        if self.closing: return
        args['fill'] = color
        try:
            x1b, y1b = self._transform(x1, y1)
            x2b, y2b = self._transform(x2, y2)
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in create_line: {(x1, y1, x2, y2)}")
        try:
            # id = self.canvas.create_line(x1b, y1b, x2b, y2b, args.copy())
            return self._canvas.create_line(*(x1b, y1b, x2b, y2b), **args)
        except Exception as _e:
            raise EasyPaintException("create_line")

    def create_text(self, x: float, y: float, text: str, font_size: int = 10,
                    anchor: str = 'center', color: str = 'black', justify: str = "left", **args):
        """Draws a line of text

        Arguments:
            x, y -- point coordinates

            text -- string

            font_size -- int (default is 10)

            anchor -- string (default is 'center')

            color -- color name (default is 'black')

            justify -- string (default is 'left')

        Returns:
            A number (identifier). You can use this 'id' to move or delete the text: erase(id)
        """
        if self.closing: return
        args['text'] = text
        args['anchor'] = anchor.lower()
        args['fill'] = color
        args['justify'] = justify
        args['font'] = ('courier', int(font_size * 1.15 + 2), 'bold')
        try:
            xb, yb = self._transform(x, y)
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in create_text: {(x, y)}")
        try:
            # return self.canvas.create_text(xb, yb, args)
            return self._canvas.create_text(*(xb, yb), **args)
        except Exception as _e:
            raise EasyPaintException("create_text")

    def erase(self, param=None):
        """Remove element from canvas

        Uses:
        \t erase() -- clean the canvas

        \t erase(*tags_or_ids): remove elements

        \t erase(tag_or_id): remove the element
        """
        if self.closing: return
        if param is None:  # Delete all elements
            try:
                for elem in self._canvas.find_all():
                    self._canvas.delete(elem)
            except tkinter.TclError:
                pass
        elif isinstance(param, list):  # Delete list of elements
            if len(param) > 0:
                try:
                    for elem in param:
                        self._canvas.delete(elem)
                except Exception as _e:
                    raise EasyPaintException(f"Wrong id in erase: {param}")
        else:  # Delete single element
            try:
                self._canvas.delete(param)
            except Exception as _e:
                raise EasyPaintException("erase")

    def save_eps(self, nombre: str):
        """Print the contents of the canvas to a postscript file.
        """
        if self.closing: return
        data = self._canvas.postscript(height=self._height, width=self._width,
                                       pagewidth='20.0c')
        try:
            f = open(nombre, 'w')
            try:
                f.write(data)
            finally:
                f.close()
            res = 1
        except Exception as _e:
            res = 0
        return res

    def move(self, tags, x: float, y: float):
        """Move items with tags
        """
        if self.closing: return
        try:
            xb = x * self._xscale
            yb = -y * self._yscale
        except Exception as _e:
            raise EasyPaintException(f"Wrong coordinates in move: x={x}, y={y}")
        try:
            self._canvas.move(tags, xb, yb)
        except Exception as _e:
            raise EasyPaintException("move")

    def close(self):
        """Terminates the program
        """
        if self.closing: return
        self.closing = True
        self._root.destroy()
        self._root.quit()

    def run(self):
        """ Launch the mainloop
        """
        if self.closing: return
        self.main()
        self._root.mainloop()

    def after(self, time: int, f: Callable[[], Any]):
        """Call function once after given time.

        Arguments:
            time -- integer that specifies the time in milliseconds

            f -- the function which shall be called
        """
        if self.closing: return
        self._root.after(time, f)

    def tag_lower(self, tag):
        """Lower an item (z-index).
        """
        if self.closing: return
        self._canvas.tag_lower(tag)

    def tag_raise(self, tag):
        """Raise an item (z-index).
        """
        if self.closing: return
        self._canvas.tag_raise(tag)

    @abstractmethod
    def main(self):
        pass


# --------------------------------------------------------------------------


if __name__ == "__main__":
    class Demo(EasyPaint):
        def on_key_press(self, keysym):
            self.close()

        def main(self):
            size = 600, 300
            #size = 'FIR'
            self.easypaint_configure(title='EasyPaint test', size=size, coordinates=(0, 300, 600, 0))
            self.create_filled_rectangle(10, 10, 590, 290, "black", "white")
            x, y = self.center
            self.create_text(x, y, "To exit press any key\nor\nclose the window", 14, justify="center")
            # self.save_eps("kk.eps")


    Demo().run()



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from random import randrange

from easypaint import EasyPaint


class Demo5(EasyPaint):
    t_id = None
    o_ids = []
    colores = ['red', 'green', 'blue', 'yellow', 'orange', 'black', 'pink']

    def dibuja_circulo_azar(self):
        x = randrange(0, 1000)
        y = randrange(0, 1000)
        tam = randrange(10, 300)
        col = self.colores[randrange(0, len(self.colores) - 1)]
        return self.create_filled_circle(x, y, tam, col)

    def on_key_press(self, keysym):
        self.close()

    def animation(self):
        if len(self.o_ids) == 100:
            self.erase(self.o_ids[0])
            del self.o_ids[0]
        self.o_ids.append(self.dibuja_circulo_azar())
        self.tag_raise(self.t_id)
        self.update()
        self.after(0, self.animation)

    def main(self):
        self.easypaint_configure(title='Demo 5 - Círculos aleatorios',
                                 background='steelblue',
                                 size=(600, 600),
                                 coordinates=(0, 0, 1000, 1000))

        self.t_id = self.create_text(500, 0, "Press any key to exit", 12, anchor="s")
        self.after(0, self.animation)


Demo5().run()



#!/usr/bin/env python3
"""
Created on 29/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
import time
from math import sin, cos, pi
from typing import *

from easypaint import EasyPaint
from easypaint.libsimple3d import Escena3D, Cubo3D, Piramide3D, Punto3D  # importa módulo de 3D

t_inc = 0.04  # incremento de t en cada frame
pi2 = 2 * pi


class Demo3(EasyPaint):
    t: float
    pv: Punto3D
    n_frames: int
    fps_id: Optional[Any]
    prev_time: float
    prev_num_lines: int
    prev_lines_ids: List[Any]

    ppd = 1500
    escena = Escena3D(ppd)

    def on_key_press(self, keysym):
        if keysym == 'Return':
            self.close()

    def animation(self):
        global t_inc
        self.n_frames += 1
        self.t = (self.t + t_inc) % pi2
        self.pv.posParametrica(self.t)
        self.escena.puntoVista(self.pv)
        lineas = self.escena.dibuja()
        num_lines = len(lineas)
        lines_ids = []
        for i in range(num_lines):
            lines_ids.append(self.create_line(*lineas[i]))
            if i < self.prev_num_lines: self.erase(self.prev_lines_ids[i])
        self.erase(self.prev_lines_ids[num_lines:])
        # if t>pi: break
        if self.t > pi - 1e-2 or self.t < 1e-2: t_inc = -t_inc
        self.prev_lines_ids = lines_ids
        self.prev_num_lines = num_lines
        if self.fps_id is not None: self.erase(self.fps_id)
        c_time = time.time()
        if c_time != self.prev_time:
            self.fps_id = self.create_text(0, -450, f"FPS: {self.n_frames / (c_time - self.prev_time):.2f}", 10, 'S')
        self.update()
        self.after(0, self.animation)

    def main(self):
        self.easypaint_configure(title='Demo 3 - Paseo 3D',
                                 background='steelblue',
                                 size=(600, 600),
                                 coordinates=(-500, -500, 500, 500))
        for x, y, z in [(1000, 1000, 1000), (1000, 1000, -1000),
                        (1000, -1000, 1000), (1000, -1000, -1000),
                        (-1000, 1000, 1000), (-1000, 1000, -1000),
                        (-1000, -1000, 1000), (-1000, -1000, -1000)]:
            if y == -1000:
                cubo = Cubo3D(1500, Punto3D(x, y, z))
            else:
                cubo = Piramide3D(Punto3D(1500, 1500, 1500), Punto3D(x, y, z))
            self.escena.insertar(cubo)
        self.create_text(0, -500, "Pulsa <Return> para salir", 10, 'S')

        self.t = 0
        self.pv = Punto3D(lambda t: 10000 * sin(t), lambda t: 10000 * sin(t), lambda t: 10000 * cos(t))
        self.n_frames = 0
        self.prev_num_lines = 0
        self.prev_lines_ids = []
        self.fps_id = None
        self.prev_time = time.time()
        self.after(0, self.animation)


Demo3().run()



#!/usr/bin/env python3
"""
Created on 29/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from easypaint import EasyPaint


class Planeta:
    def __init__(self, x, y, vx, vy, masa, color):
        self.x, self.y, self.vx, self.vy, self.masa, self.color = x, y, vx, vy, masa, color

    def dibuja(self, ec):
        return ec.create_circle(self.x, self.y, self.masa, self.color)


def gravedad(p1, lp):
    for p2 in lp:
        x21 = p2.x - p1.x
        y21 = p2.y - p1.y
        x12 = p1.x - p2.x
        y12 = p1.y - p2.y
        r3 = (x21 * x21 + y21 * y21) ** 1.5
        masa1r3 = p1.masa / r3
        masa2r3 = p2.masa / r3
        p1.vx += masa2r3 * x21
        p1.vy += masa2r3 * y21
        p2.vx += masa1r3 * x12
        p2.vy += masa1r3 * y12
    p1.x += p1.vx
    p1.y += p1.vy


class Demo2(EasyPaint):
    planets_ids = []
    planets = []

    def on_key_press(self, keysym):
        self.close()

    def animation(self, c):
        old_x = []
        old_y = []
        for p in self.planets:
            old_x.append(p.x)
            old_y.append(p.y)
        for _ in range(15):
            for i in range(len(self.planets)):
                gravedad(self.planets[i], self.planets[i + 1:])
        old_lc, self.planets_ids = self.planets_ids, []
        for i in range(len(self.planets)):
            self.planets_ids.append(self.planets[i].dibuja(self))
            self.erase(old_lc[i])
            self.create_line(old_x[i], old_y[i], self.planets[i].x, self.planets[i].y, self.planets[i].color)
        self.update()
        if c > 0:
            self.after(5, lambda: self.animation(c - 1))

    def main(self):
        t = 400
        self.easypaint_configure(title='Demo 2 - Caos gravitatorio',
                                 background='black',
                                 size=(600, 600),
                                 coordinates=(-t, -t, t, t))

        # Si quieres añadir mas planetas, adelante...
        self.planets.append(Planeta(-200.0, -200.0, 0.1, 0.0, 20.0, 'red'))
        self.planets.append(Planeta(200.0, 200.0, -0.1, 0.0, 20.0, 'magenta'))
        self.planets.append(Planeta(0.0, 0.0, 0.1, 0.1, 0.01, 'green'))

        for p in self.planets: self.planets_ids.append(p.dibuja(self))

        self.create_text(0, -400, "Press any key to exit", 10, 'S', 'white')
        self.after(0, lambda: self.animation(1000))


Demo2().run()



#!/usr/bin/env python3
"""
Created on 28/09/2021

@author: David Llorens
@contact: dllorens@lsi.uji.es
@copyright: Universitat Jaume I de Castelló (2021)
"""
from enum import Enum

from easypaint import EasyPaint


class State(Enum):
    ToDelete = 1
    ToExit = 2


class Demo1(EasyPaint):
    state: State = State.ToDelete

    def on_key_press(self, keysym):
        if self.state == State.ToDelete:
            self.erase()
            self.create_text(250, 250, "Press any key to exit", 12)
            self.state = State.ToExit
        elif self.state == State.ToExit:
            self.close()

    def main(self):
        self.easypaint_configure(title='Demo 1 - Funciones predefinidas',
                                 size=(501, 501),
                                 coordinates=(0, 0, 500, 500))
        o_ids = []
        # Dibuja matriz de puntos
        for x in range(25, 225, 21):
            for y in range(275, 475, 21):
                o_ids.append(self.create_point(x, y, ['black', 'red'][(x + y) % 2]))

        # Dibuja dos circulos (uno relleno y otro no)
        o_ids.append(self.create_filled_circle(375, 375, 125, 'black', 'blue'))
        o_ids.append(self.create_circle(125, 125, 125, 'red'))

        # Dibuja dos rectangulos (uno relleno y otro no)
        o_ids.append(self.create_rectangle(150, 75, 200, 175, 'red'))
        o_ids.append(self.create_filled_rectangle(50, 75, 100, 175, 'black', 'red'))

        # Dibuja dos lineas en cruz
        o_ids.append(self.create_line(250, 125, 500, 125, 'green'))
        o_ids.append(self.create_line(375, 0, 375, 250, 'black'))

        # escribe texto
        o_ids.append(self.create_text(250, 250, "Press any key to delete all", 12))


Demo1().run()
